/* ********************************************************************** *
 * Global Window
 * ********************************************************************** */

interface Window {
  readonly imHichee: IImHichee;
}

interface IImHichee {
  /** Element Repository */
  UIElementRepository: IUIElementRepository;

  /** Container Action Items Repository */
  UIContainerActionItemRepository: IUIContainerActionItemRepository;

  /** Container Action Condition Repository */
  UIContainerActionConditionRepository: IUIContainerActionConditionRepository;

  /** Category Repository of Component Properties */
  UIComponentPropertyCategoryRepository: IUIComponentPropertyCategoryRepository;

  /** HTML Element Builder */
  HTMLElementBuilder: {
    /** Create instantiate from element */
    fromElement: (tag: HTMLElement) => IHTMLElementBuilder;

    /** Create instantiate from Tag Name */
    createElement: (tagName: string, clazz: new () => HTMLElement) => IHTMLElementBuilder;
  };

  /** Browser Helper */
  BrowserHelper: {
    /**
     * Browser Identification
     *
     * Get information about the system and browser on which the container is running. \
     * IM-BloomMaker itself uses this property to determine the browser.
     */
    readonly browser: BrowserInformation;

    /**
     * Execute callback at next draw time
     *
     * Release control of the current thread and execute a callback earlier than `setTimeout`.
     */
    nextTick: (callback: () => void) => void;

    /**
     * fire resize event
     *
     * Prompt the designer or element to recalculate or re-render the size by firing the resize event manually.
     */
    dispatchResizeEvent: () => void;

    /**
     * Safely Executable `setTimeout`
     *
     * special `setTimeout` to prevent getting `window`. \
     * The first argument cannot be a string.
     */
    safeSetTimeout: () => number;

    /**
     * Safely Executable `setInterval`
     *
     * special `setTimeout` to prevent getting `window`. \
     * The first argument cannot be a string.
     */
    safeSetInterval: () => number;

    /**
     * Get BASE URL
     *
     * Gets the BASE URL for the current page.
     */
    getBaseURL: () => string;

    /**
     * Display Alert Dialog
     *
     * Display the original alert dialog provided by IM-BloomMaker itself. \
     * This dialog box does not interfere with the closing of the browser window or the behavior of other JavaScript. \
     * Therefore, if this method is called, `await` or if wait processing is not implemented, subsequent processing will continue.
     *
     * This method returns Promise. \
     * If the "OK" button is clicked in the dialog, the promise is followed.
     */
    alert: (message?: CSJSPrimitiveAllType) => Promise<void>;

    /**
     * Display Confirmation Dialog
     *
     * Display the original confirmation dialog provided by IM-BloomMaker itself. \
     * This dialog box does not interfere with the closing of the browser window or the behavior of other JavaScript. \
     * Therefore, if this method is called, `await` or if wait processing is not implemented, subsequent processing will continue.
     *
     * This method returns Promise. \
     * If the "OK" or "Cancel" button is clicked in the dialog, subsequent processing of the Promise is followed. \
     * The "OK" button returns `true`, the "Cancel" button returns `false` asynchronously.
     */
    confirm: (message?: CSJSPrimitiveAllType) => Promise<boolean>;

    /**
     * Display Error Dialog
     *
     * Display the original error dialog provided by IM-BloomMaker itself. \
     * This dialog box does not interfere with the closing of the browser window or the behavior of other JavaScript. \
     * Therefore, if this method is called, `await` or if wait processing is not implemented, subsequent processing will continue.
     *
     * This method returns Promise.. \
     * If the "OK" button is clicked in the dialog, the promise is followed.
     */
    showErrorDialog: (message?: CSJSPrimitiveAllType) => Promise<void>;
  };

  /** Argument Factory */
  ArgumentFactory: {
    /** Create ArrayArgument */
    createAsArray: (container: IUIContainer, raw: IArgument[]) => IArrayArgument;

    /** Create BooleanArgument */
    createAsBoolean: (container: IUIContainer, raw: boolean) => IArgument;

    /** Create DateArgument */
    createAsDate: (container: IUIContainer, raw: string | number | {date: Date; offset?: number}) => IDateArgument;

    /** Create DoubleArgument */
    createAsDouble: (container: IUIContainer, raw: number) => INumberArgument;

    /** Create FractionArgument */
    createAsFraction: (container: IUIContainer, raw: IFraction) => INumberArgument;

    /** Create IntegerArgument */
    createAsInteger: (container: IUIContainer, raw: number) => INumberArgument;

    /** Create NullArgument */
    createAsNull: (container: IUIContainer) => IArgument;

    /** Create ObjectArgument */
    createAsObject: (container: IUIContainer, raw: IKeyValueMap<string, IArgument>) => IObjectArgument;

    /** Create ParameterPathArgument */
    createAsParameterPath: (container: IUIContainer, target: IUIComponent, raw: string) => IParameterPathArgument;

    /** Create StringArgument */
    createAsString: (container: IUIContainer, raw: string) => IArgument;
  };

  /** Numbering Element ID */
  UIElementIDResolver: IUIElementIDResolver;

  /** Message Builder */
  MessageBuilder: IMessageBuilder;

  /** UUID */
  UUID: {
    /** Create UUID */
    generate: () => string;
  };

  /** Designer Content Information */
  content: {
    type: string;
    contentId: string;
    categoryId: string;
    cssFramework: string;
    name: string;
    localizedName: {[key: string]: string};
    description: string;
    localizedDescription: {[key: string]: string};
    sourceTemplateId: string;
    sortNumber: number;
    version: string;
    plugins: {id: string; [key: string]: CSJSPrimitiveSingleType}[];
    token: string;
  };

  /** Designer Template Information */
  template: {
    type: string;
    templateId: string;
    categoryId: string;
    cssFramework: string;
    name: string;
    localizedName: {[key: string]: string};
    description: string;
    localizedDescription: {[key: string]: string};
    sourceContentId: string;
    sortNumber: number;
    plugins: {id: string; [key: string]: CSJSPrimitiveSingleType}[];
    token: string;
  };

  /** Setting */
  configuration: {
    /** for Layout */
    layout: {
      /** Maximum value of elementCount property */
      maxElementCount: number;
      /** Maximum value of rowCount property */
      maxRowCount: number;
      /** Maximum value of columnCount property */
      maxColumnCount: number;
    };
    /** for rendering */
    render: {
      /** Maximum time of Repeating Element Rendering (ms) */
      maxRepeatableCreateTime: number;
    };
  };

  /** Context from server */
  contexts: {
    /** Account Context */
    accountContext: {
      /** Locale ID for multilingual locale resolution */
      localeId: string;
      /** Time zone for DateTime */
      timeZone: {
        /** TimeZoneID */
        id: string;
        /** Offset from UTC (milliseconds) */
        rawOffset: number;
        /** Daylight Saving Time */
        useDaylightTime: boolean;
      };
    };
  };

  /** System information from the server */
  systemInfo: {
    /** System Information */
    system: {
      /** Locale ID */
      localeId: string;

      /** Locales available on the system */
      availableLocales: {
        /** Locale ID */
        localeId: string;
        /** locale display name */
        displayName: string;
        /** Default encoding */
        defaultEncoding: string;
      }[];
    };

    /** Tenant Information */
    tenant: {
      /** Locale ID */
      localeId: string;
    };
  };

  /** Content Identification (Content or Template) */
  readonly contentDefinitionType: string;

  /** ImSecureToken */
  readonly imSecureToken: string;

  /** The "Input" of the variable. Used only in execution mode. */
  readonly inputData?: string;

  /** Content data. Used only in execution mode. */
  readonly serializedData?: string;
}

/* ********************************************************************** *
 * Component
 * ********************************************************************** */

interface IUIComponent {
  /**
   * Component Class
   *
   * Gets the class that makes up the component.
   */
  readonly clazz: UIComponentClass;

  /**
   * Component Placement Constraints
   *
   * Gets the restrictions on element copying, deleting, placing and so on.
   */
  readonly constraint: IUIComponentConstraint;

  /**
   * Component Properties
   *
   * Gets the raw value set for the property this component has.
   */
  readonly properties: IUIComponentProperties;

  /**
   * enabled state
   *
   * Returns true if the component is not being destroy after it is instantiated.
   * Returns false if being in the unavailable state on the container due to the destroy method and so on.
   */
  readonly alived: boolean;

  /**
   * Rendering State
   *
   * Returns true if rendered on the container (be displayed on the screen).
   * Returns `false `if not rendered.
   */
  readonly rendered: boolean;

  /**
   * Wrapped HTML Element Builder for rendering
   *
   * Gets the Element Builder for constructing the root tag when the component is actually rendered. \
   * At render time, this Element Builder is used to build the actual node.
   */
  readonly wrappedBuilder: IHTMLElementBuilder;

  /**
   * Array of child elements
   *
   * Gets the child components under this component. \
   * The order of components returned by this property is not guaranteed. \
   * Use the `IUIElementController#getSortedChildren` method to get them in order in which they are rendered on the container.
   */
  readonly children: IUIElement[];

  /** UUID assigned per component */
  readonly uuid: string;

  /** component display location */
  getClientRect(): UIComponentClientRect;

  /** Internal position of component */
  getOffsetRect(): UIComponentClientRect;
}

type UIComponentClass = 'UIContainer' | 'UIElement';

type UIComponentClientRect = {
  left: number; // Rewritable as it being categorized as a temporary variable
  right: number; // Rewritable as it being categorized as a temporary variable
  top: number; // Rewritable as it being categorized as a temporary variable
  bottom: number; // Rewritable as it being categorized as a temporary variable
  width: number; // Rewritable as it being categorized as a temporary variable
  height: number; // Rewritable as it being categorized as a temporary variable
};

interface IUIComponentConstraint {
  /**
   * Movability
   *
   * Returns `true` if the component can be moved in the designer. \
   * If not, returns `false`. \
   * Regardless of this return value, the component can be moved when being operated via the API.
   */
  readonly movable: boolean;

  /**
   * Deletability
   *
   * Returns `true` if the component can be deleted in the designer. \
   * If not, returns `false`. \
   * Regardless of this return value, the component can be deleted when being operated via the API.
   */
  readonly removable: boolean;

  /**
   * Copyability
   *
   * Returns `true` if the component can be duplicated in the designer. \
   * Returns `false `if it cannot be duplicated. \
   * Regardless of this return value, the component can be duplicated when being operated via the API.
   */
  readonly copyable: boolean;

  /**
   * end flag
   *
   * Returns `true` if the element cannot be placed below this one. \
   * Returns `false` if child elements can be placed.
   *
   * This means that `true` is returned if the `acceptableChild` method always returns `false`.
   * false must be returned in the other cases.
   */
  readonly extremity: boolean;

  /**
   * Get Parental Acceptability as Child
   *
   * Returns `true` if the element passed in the argument is accepted as a parent. \
   * Otherwise, it returns `false`.
   */
  acceptableParent(parent: ParentUIElement): boolean;

  /**
   * Get child acceptability as parent
   *
   * Returns `true` if the element passed in the argument is accepted as a child. \
   * Otherwise, it returns `false`.
   */
  acceptableChild(child: IUIElement): boolean;
}

interface IUIComponentProperties {
  /** Get Properties */
  getProperty(key: string): IUIComponentProperty;

  /** enumerate properties */
  forEach(callback: UIComponentPropertiesCallback): void;

  /** check for key presence */
  containsKey(key: string): boolean;
}

type UIComponentPropertiesCallback = (key: string, property: IUIComponentProperty) => void;

interface IUIComponentProperty {
  /** Property Key */
  readonly key: string;

  /** Display Name */
  readonly displayName: string;

  /** Property Definition */
  readonly definition: IUIComponentPropertyDefinition;

  /** Input Type */
  readonly type: UIComponentPropertyType;

  /** Property Value */
  readonly value: IPropertyValueArgument;
}

type UIComponentPropertyType = 'string' | 'integer' | 'decimal' | 'boolean' | 'none' | 'action';

/**
 * Component property descriptions are specified in the message property key of `SIT.Z.IWP.HICHEE.COMPONENT.PROPERTY.DESCRIPTION. ${Component Name}. ${property key} `.
 * If the message property key does not exist, the help icon for the property is not displayed.
 */
interface IUIComponentPropertyDefinition {
  /** Required Items */
  readonly required: boolean;

  /** Type of data */
  readonly dataType: UIComponentPropertyDataType;

  /** Minimum allowed value */
  readonly min: number | null;

  /** Maximum allowed value */
  readonly max: number | null;

  /** Read Only (User not rewritable) */
  readonly readonly: boolean;

  /** Re-render on value change */
  readonly rerender: boolean;

  /** Repeated Element Use */
  readonly repeatable: boolean;

  /** Validation execution during preview and on execution screen */
  readonly validateValue: boolean;

  /** Absolute value is allowed to be specified */
  readonly acceptStatic: boolean;

  /** Relative value is allowed to be specified */
  readonly acceptDynamic: boolean;

  /** Multi-line strings is allowed to be input */
  readonly acceptMultipleLines: boolean;

  /** Input Candidates */
  readonly candidate: IUIComponentPropertyCandidate[] | null;
}

type UIComponentPropertyDataType = 'NUMERIC' /** fixed-point */ | 'INTEGER' /** integer */ | 'ANY' /** unspecified */;

interface IUIComponentPropertyCandidate {
  /** Display Name */
  readonly displayName: string;

  /** Actual Value */
  readonly value: UIComponentPropertyValueType;
}

type UIComponentPropertyValueType = string | number | boolean | null;

interface IUIComponentPropertyCategoryRepository {
  /** List of Property Categories */
  readonly categories: UIComponentPropertyCategoryInfo[];

  /** Register a property category */
  registerCategory(categoryInfo: UIComponentPropertyCategoryInfo): void;
}

/**
 * The property category name is specified in the message property key of `CAP.Z.IWP.HICHEE.COMPONENT.PROPERTY.CATEGORY.NAME. {Category ID} `.
 * If this key does not exist, the category ID is used.
 */
type UIComponentPropertyCategoryInfo = {
  /** Category ID */
  readonly id: string;

  /** Sort Number (ascending order) */
  readonly sortNumber?: number;

  /** List of display types to be hidden display conditions */
  readonly hiddenDisplayTypes?: IUIComponentDisplayType[];
};

interface IUIComponentDisplayType {
  /** class name */
  readonly className: string;

  /** Auto width adjustment flag */
  readonly autoWidth: boolean;

  /** Auto Height Adjustment Flag */
  readonly autoHeight: boolean;
}

type ParentUIElement = IUIContainer | IUIElement;
type NextUIElement = IUIElement | null;

type UIComponentConstraintClass = new () => IUIComponentConstraint;

/* ********************************************************************** *
 * Container
 * ********************************************************************** */

interface IUIContainer extends IUIComponent {
  /**
   * Designer Mode
   *
   * Returns `true` if the container is to be used on the designer screen. \
   * Returns `false` for containers used on the preview and execution screens.
   */
  readonly designMode: boolean;

  /**
   * working mode
   *
   * Returns `true` for containers used for non-display purposes, such as for clipboard purposes. \
   * Returns `false` if the container is to be used for on-screen display.
   */
  readonly workingMode: boolean;

  /**
   * elementary repository
   *
   * Gets the repository that manages the elements available in this container.
   */
  readonly repository: IUIElementRepository;

  /**
   * Element Controllers
   *
   * Gets the controller that manages instances of elements added to this container.
   */
  readonly controller: IUIElementController;

  /**
   * parameter storage
   *
   * Retrieves storage that manages variables, constants, inputs, and multilingual information that are valid in this container.
   */
  readonly parameters: IParameterStorage;

  /**
   * action chart storage
   *
   * Gets the storage that manages action charts that summarize the action items available in this container.
   */
  readonly eventCharts: IActionChartStorage;

  /**
   * Container
   *
   * Get itself.
   */
  readonly container: IUIContainer;

  /**
   * Component Properties
   *
   * Gets the raw value set for the property this container has.
   */
  readonly properties: IUIContainerProperties;

  /**
   * scroll position
   *
   * If a container size is limited, get the scroll position of the inner area of the container.
   */
  readonly scroll: {left: number; top: number};

  /**
   * Validation error presence flag
   *
   * Gets whether validation errors exist for all elements in the container. \
   * If there are any validation errors, `true` is returned.
   */
  readonly hasValidationError: boolean;

  /**
   * Rendering
   *
   * Initializes each event driver in the container and renders the element it is associated with. \
   * Restarting after rendering causes re-rendering.
   */
  render(force?: boolean): void;

  /**
   * Deletion
   *
   * When you delete an element, the child elements are automatically deleted and the parent-child relationship is broken. \
   * If you delete a component, you cannot restore it.
   */
  destroy(): void;

  /**
   * loading display
   *
   * Displays the loading layer in the container during the asynchronous operation specified in the argument. \
   * When asynchronous processing is finished, the loading layer is automatically deleted.
   */
  showLoadingIndicator<T>(promise: Promise<T>): Promise<T>;
}

interface IUIContainerProperties extends IUIComponentProperties {}

/* ********************************************************************** *
 * Element
 * ********************************************************************** */

interface IUIElement extends IUIComponent {
  /**
   * Display Type Name
   *
   * Gets a string that indicates how it is displayed on the browser screen.
   */
  readonly displayType: IUIComponentDisplayType;

  /**
   * Container
   *
   * Gets the container this element belongs to.
   */
  readonly container: IUIContainer;

  /**
   * Component Placement Constraints
   *
   * Gets the restrictions on copying, deleting, placing, etc. of elements.
   */
  readonly constraint: IUIComponentConstraint;

  /**
   * Component Properties
   *
   * Gets the raw value set for the property this element has. \
   * Use `propertyAccessor` to get the final resolved value that takes into account fixed and variable values on the execution screen.
   */
  readonly properties: IUIElementProperties;

  /**
   * Property Accessors
   *
   * Provides the ability to get and set new values for Arguments that are set in component properties. \
   * When a value is retrieved, the final resolved value considering fixed and variable values is returned.
   */
  readonly propertyAccessor: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>;

  /**
   * Element instance
   *
   * Gets the instance of the actual class that defines the behavior or processing of the element.
   */
  readonly instance: IUIElementCore;

  /**
   * Source HTML Element Builder
   *
   * Gets the Element Builder for constructing the tag that is the root of the element. \
   * At rendering, this Element Builder is used to build the actual node.
   */
  readonly builder: IHTMLElementBuilder;

  /**
   * Parent Element
   *
   * Gets the parent element or container where this element is located. \
   * Rewriting this property will move the element to the end of its subordinates, with the assigned element or container as the parent.
   */
  parent: ParentUIElement;

  /**
   * Next Element
   *
   * Gets the element next to this element. \
   * Returns `null` if it is the last element.
   *
   * Rewriting this property will move the element in front of the assigned element. \
   * If you assign null, move the element to the end of its parent.
   */
  next: NextUIElement;

  /**
   * Next rendered element
   *
   * Get the previously rendered element next to this element. \
   * Returns `null` if it is the last element or there is no rendered element next.
   */
  readonly renderedNext: NextUIElement;

  /**
   * Display Status
   *
   * Get or set whether to place the node on the screen. \
   * Set `true` if you want the element to appear on the screen when placed on the node tree. \
   * Set `false` to remove the element from the node tree and hide it from the screen.
   *
   * The initial state is `true`.
   */
  visibility: boolean;

  /**
   * Rendering
   *
   * In the delete state, nothing is done. \
   * Use the `resume` method to release the delete state and re-rendering.
   */
  render(force?: boolean): void;

  /**
   * Deletion
   *
   * When you delete an element, the child elements are automatically deleted and the parent-child relationship is broken. \
   * Using the `resume `method will not restore child elements.
   */
  destroy(): void;

  /**
   * Resume Delete
   *
   * The `destroy `method removes the deletion and restores the instance. \
   * However, child elements are not restored.
   */
  resume(): void;

  /**
   * Duplicate Instance
   *
   * Duplicates an instance of the element to the specified parent and returns the new element. \
   * Child elements are also automatically duplicated.
   */
  clone(parent?: ParentUIElement): IUIElement;

  /**
   * Perform variable value validation of properties
   *
   * Compares the variable value validation specified for each property of an element with the value entered for the property, and returns the result. \
   * If at least one property violates the validation rule, it returns `true` to display a red border on the execution screen. \
   * Always run asynchronously, so implement `await` or wait processing appropriately.
   */
  validate(): Promise<boolean>;
}

interface ISimpleUIElement extends IUIElement {}

interface IIndexableUIElement extends IUIElement {
  /**
   * Repeating index
   *
   * Get or set the position of the element relative to its parent. \
   * This index is used by `$index` in the variable path.
   *
   * This value is overwritten by IM-BloomMaker itself during element rendering. \
   * Do not overwrite values from the element side.
   */
  repeatIndex: number;
}

interface IRepeatableUIElement extends IIndexableUIElement {}

interface IUIElementCore {
  /**
   * Element Type Name
   *
   * Gets a string that is identical to the element's class name.
   */
  readonly elementTypeName: string;

  /**
   * Display Type Name
   *
   * Gets a string that indicates how it is displayed on the browser screen.
   */
  readonly displayType: UIElementDisplayType;

  /**
   * Constraint class of the element
   *
   * Gets the class that contains restrictions on copying, deleting, positioning, etc. of elements.
   */
  readonly constraintClass: UIComponentConstraintClass;

  /**
   * Element-specific property definition
   *
   * Gets the unique property definition of the element. \
   * You can set any key, but you cannot specify a key that duplicates `commonPropertyDefinition`.
   */
  readonly uniquePropertyDefinition: IUniquePropertyDefinition;

  /**
   * Element Common Property Definitions
   *
   * Gets the common property definition of all elements.
   */
  readonly commonPropertyDefinition: ICommonPropertyDefinition;

  /**
   * Dependent Element Classes
   *
   * If `createChildren`, `updateChildren`, `cloneChildren` uses elements other than itself as child elements,
   * When you return the element's class, IM-Hichee automatically registers the class and makes it available for use.
   */
  readonly dependElements: UIElementCoreClass[];

  /**
   * Element wrapper class
   *
   * To create a simple and non-repeating element, return `SimpleUIElement`,
   * to create a manually repeating element, return `IndexableUIElement`,
   * to create an automatically repeating element, return `RepeatableUIElement`.
   */
  readonly wrapperClass: UIElementWrapperClass;

  /**
   * Create HTML Element
   *
   * This method is called only once by the IM-Hichee side when a virtual element is created. \
   * Create an instance of the root tag `IHTMLElementBuilder` and return it. \
   * Instances of `IHTMLElementBuilder`can be found in `window.imHichee.HTMLElementBuilder.createElement`
   * or `window.imHichee.HTMLElementBuilder.fromElement` method.
   *
   * Fort the instance of `IHTMLElementBuilder` that returns in this method,
   * do not use the setCSS, removeCSS, or switchCSS methods. \
   * Note that there are no restrictions on child element instances.
   *
   * The `build` method of `IHTMLElementBuilder` is called automatically and does not need to be called explicitly.
   */
  createElement(
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>
  ): IHTMLElementBuilder;

  /**
   * Update HTML Element
   *
   * This method is called by IM-Hichee whenever rendering is needed. \
   * Perform the necessary rendering on the instance of `IHTMLElementBuilder` passed in the argument.
   *
   * The `build` method of `IHTMLElementBuilder` is called automatically and does not need to be called explicitly.
   */
  updateElement(
    builder: IHTMLElementBuilder,
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>
  ): void;

  /**
   * Delete HTML Element
   *
   * Method called once by IM-Hichee when a virtual element is deleted. \
   * Perform the required deallocation on the instance of `IHTMLElementBuilder` passed in the argument.
   *
   * The `tag` registered with `IHTMLElementBuilder` is automatically deparented, so there is no need to explicitly untag it.
   */
  destroyElement(
    builder: IHTMLElementBuilder,
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>
  ): void;

  /**
   * Create Child Element
   *
   * This method is called only once by the IM-Hichee side when the virtual element is purely created (without (any) cloning). \
   * Return an instance of the child element that should be automatically prepared when this element is created.
   *
   * The child elements are arranged in the order in which they are stored in the array.
   */
  createChildren(
    self: IUIElement,
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>
  ): IUIElement[];

  /**
   * Update Child Elements
   *
   * This method is called by IM-Hichee whenever rendering is needed. \
   * If there is a change in the array of child elements passed in the argument, return the modified array.
   *
   * If there are no changes, simply return the argument `children`. \
   * The child elements are arranged in the order in which they are stored in the array.
   */
  updateChildren(
    children: IUIElement[],
    self: IUIElement,
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>
  ): IUIElement[];

  /**
   * Cloned Child Element
   *
   * This method is called only once by the IM-Hichee side when a virtual element is created by a clone. \
   * An array of cloned child elements is passed as an argument. The child element `parent`, `next-` is undefined at this point.
   *
   * If you have stored the child element in the instance, you should store the new instance again in this method.
   *
   * If there are no changes, simply return the argument `children`. \
   * The child elements are arranged in the order in which they are stored in the array.
   */
  clonedChildren(
    children: IUIElement[],
    self: IUIElement,
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>
  ): IUIElement[];
}

type UIElementDisplayType =
  | 'BLOCK' /** Block Element */
  | 'INLINE' /** Inline Block Element */
  | 'INLINE_FLEX' /** Inline Flex Element */
  | 'INPUT' /** Inline Input Element */
  | 'SEPARATOR' /** Separator Element */
  | 'TABLE' /** Table Element */
  | 'TABLE_CELL' /** Table Cell Element */
  | 'FLEX_BOX' /** Flex Element */
  | 'FLEX_ITEM' /** Flex Item Element */
  | 'FIXED' /** Undesigned Element */
  | 'SELECTABLE_FIXED' /** Undesigned Selectable Elements */
  | 'FUNCTIONAL' /** Functional Undesigned Elements */
  | 'CANVAS' /** Inline Elements With Limited Functionality */
  | 'PURE'; /** Elements Without display */

type UIElementCoreClass = new () => IUIElementCore;
type UIElementWrapperClass =
  | 'IndexableUIElement' /** Element With Index */
  | 'RepeatableUIElement' /** Repeating Element */
  | 'SimpleUIElement'; /** Simple Regular Elements */

interface IIndexableUIElementCore extends IUIElementCore {}

interface IRepeatableUIElementCore extends IUIElementCore {
  /**
   * Repeat Count
   *
   * Returns the number of times the child element is repeated. \
   * The child elements are automatically duplicated for the number of times returned here.
   */
  repeatCount(properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>): number;
}

interface IUIElementProperties extends IUIComponentProperties {
  /**
   * Element ID
   *
   * A unique ID that identifies an element on the screen. \
   * This allows you to change the element's ID attribute, but it does not perform duplicate checking, so be careful not to duplicate it.
   */
  elementId: string;
}

interface IUIElementPropertyAccessor<DEFINITION> {
  /** Get Properties */
  getProperty<K extends keyof DEFINITION>(key: K): IArgument;

  /**
   * Set Property Value
   *
   * Specify the type other than `IArgument` to `value` and the property is a variable value, assign it to the variable value. \
   * Otherwise, change the value of the property itself.
   *
   * If the value specified for `value` is indeterminate (Input in progress, etc.), specify `fixed` as `false`. \
   * Specify `fixed` as `true` if confirmed.
   */
  setProperty<K extends keyof DEFINITION>(key: K, value: CSJSPrimitiveAllType, fixed: boolean): void;
}

interface IUniquePropertyDefinition {
  [key: string]: UniquePropertyDefinitionType;
}

/**
 * Component property descriptions are specified in the message property key of `SIT.Z.IWP.HICHEE.COMPONENT.PROPERTY.DESCRIPTION. ${Component Name}. ${property key} `.
 * If the message property key does not exist, the help icon for the property is not displayed.
 */
type UniquePropertyDefinitionType = {
  /** Display Name */
  readonly displayName: string;

  /** Optional definition */
  readonly definition: {
    /** Validation Rule - Required (If unspecified: `false`) */
    readonly required?: boolean;

    /** Validation Rule - Data Type (If unspecified: `ANY`) */
    readonly dataType?: UIComponentPropertyDataType;

    /** Validation Rule - Minimum Number */
    readonly min?: number;

    /** Validation Rule - Maximum Number */
    readonly max?: number;

    /** Read Only (If unspecified: `false`) */
    readonly readonly?: boolean;

    /** Re-rendering on value change (If unspecified: false) */
    readonly render?: boolean;

    /** Repeated use of elements (If unspecified: `false`) */
    readonly repeatable?: boolean;

    /** Validation execution during preview and execution screen (If unspecified: `false`) */
    readonly validateValue?: boolean;

    /** Allow Fixed Values (If unspecified: `true`) */
    readonly acceptStatic?: boolean;

    /** Allow Variable Values (If unspecified: `true`) */
    readonly acceptDynamic?: boolean;

    /** Allow Multiline Text (If unspecified: `false`) */
    readonly acceptMultipleLines?: boolean;

    /** Value choices */
    readonly candidate?: {
      /** Display Name */
      readonly displayName: string;

      /** Value */
      readonly value: UIComponentPropertyValueType;
    }[];
  };

  /** Property Type Name */
  readonly type: UIComponentPropertyType;

  /** Property Value */
  readonly value?: {};

  /** Category ID */
  readonly categoryId?: string;
};

interface ICommonPropertyDefinition {
  /** onMouseDown Event */
  readonly onMouseDown?: EventPropertyDefinitionValueType;

  /** onMouseMove event */
  readonly onMouseMove?: EventPropertyDefinitionValueType;

  /** onMouseLeave Event */
  readonly onMouseLeave?: EventPropertyDefinitionValueType;

  /** onMouseOut event */
  readonly onMouseOut?: EventPropertyDefinitionValueType;

  /** onMouseOver Event */
  readonly onMouseOver?: EventPropertyDefinitionValueType;

  /** onMouseUp Event */
  readonly onMouseUp?: EventPropertyDefinitionValueType;

  /** onKeyDown Event */
  readonly onKeyDown?: EventPropertyDefinitionValueType;

  /** onKeyPress event */
  readonly onKeyPress?: EventPropertyDefinitionValueType;

  /** onKeyUp Event */
  readonly onKeyUp?: EventPropertyDefinitionValueType;

  /** onFocus Event */
  readonly onFocus?: EventPropertyDefinitionValueType;

  /** onBlur Event */
  readonly onBlur?: EventPropertyDefinitionValueType;

  /** onChange Event */
  readonly onChange?: EventPropertyDefinitionValueType;

  /** onInput Event */
  readonly onInput?: EventPropertyDefinitionValueType;
}

type EventPropertyDefinitionValueType = 'use' /** Use this property */;

interface IUIElementConstructorArguments {
  /** Container */
  readonly container: IUIContainer;

  /** Parent element */
  readonly parent: ParentUIElement;

  /** Next Element */
  readonly next: NextUIElement;

  /** UUID */
  readonly uuid: string | null;

  /** Element ID */
  readonly elementId: string | null;

  /** Original element */
  readonly source: IUIElement | null;

  /** Allay of child elements */
  readonly children: IUIElement[] | null;

  /** Auto Clone Child Elements */
  readonly cloneChildren: boolean;
}

interface IUIElementRepository {
  /** List of element classes */
  readonly classes: UIElementCoreClass[];

  /** List of element categories */
  readonly categories: UIElementCategoryInfo[];

  /** Register Element Class */
  registerClass(clazz: UIElementCoreClass, options?: UIElementRegisterOptions): void;

  /** Register element categories */
  registerCategory(categoryId: string, options?: UIElementCategoryRegisterOptions): void;

  /** Get Element Class */
  getClass(elementTypeName: string): UIElementCoreClass;

  /** Get options for element class */
  getClassOptions(clzz: UIElementCoreClass): UIElementRegisterOptions;

  /** Set Element Class Options */
  setClassOptions(clazz: UIElementCoreClass, options: UIElementRegisterOptions): void;

  /** Get options for element category */
  getCategoryOptions(categoryId: string): UIElementCategoryInfo;

  /** Set element category options */
  setCategoryOptions(categoryId: string, options: UIElementCategoryRegisterOptions): void;
}

/**
 * The component category name is specified in the message property key of `CAP.Z.IWP.HICHEE.COMPONENT.CATEGORY.NAME. {Category ID} `.
 * If this key does not exist, use category ID.
 */
type UIElementCategoryInfo = {
  /** Category ID */
  readonly categoryId: string;

  /** Options */
  readonly options?: UIElementCategoryRegisterOptions;
};

type UIElementCategoryRegisterOptions = {
  /** Sort Number (ascending order) */
  readonly sortNumber?: number;
};

/**
 * The component name is specified in the message property key of `CAP.Z.IWP.HICHEE.COMPONENT.NAME. {Class Name} `.
 * If this key does not exist, use the class name.
 */
type UIElementRegisterOptions = {
  /** Category ID */
  readonly categoryId?: string;

  /** Sort Number */
  readonly sortNumber?: number;

  /**
   * Hide flag in palette
   *
   * This flag is overwritten by the palette in the designer.
   */
  readonly hiddenInPalette?: boolean;
};

interface IUIElementIDResolver {
  /** Element ID Numbering */
  createElementId(): string;

  /** Get Element ID Branch Number */
  createElementIdWithSuffixCode(element: IUIElement): string;

  /**
   * Register Element ID
   *
   * If there is a conflict, the new element ID is returned.
   */
  registerElementId(element: IUIElement): string;

  /** Free Element ID */
  releaseElementId(element: IUIElement): void;

  /** Clear */
  clear(): void;
}

interface IUIElementController {
  /**
   * array of instances
   *
   * Get all elements managed by this element controller.
   */
  readonly instances: IUIElement[];

  /**
   * Create an instance of an element
   *
   * Creates an instance of an element by specifying the `elementTypeName` that the element has.
   */
  createInstanceByTypeName(elementTypeName: string, parent: ParentUIElement, next?: NextUIElement): IUIElement;

  /**
   * Create an instance of an element
   *
   * Specifies the element class and creates an instance of the element.
   */
  createInstanceByClass<T extends IUIElement>(
    clazz: UIElementCoreClass,
    parent: ParentUIElement,
    next?: NextUIElement
  ): T;

  /**
   * Duplicate an instance of an element
   *
   * Duplicates an instance of an element. \
   * `parent` allows you to specify the destination of the replication. If not specified, duplicates to the same parent as the source. \
   * If `cloneChildren` is `true`, the child elements are also recursively duplicated.
   */
  cloneInstance(source: IUIElement, parent?: ParentUIElement, cloneChildren?: boolean): IUIElement;

  /**
   * Deserialize an instance of an element
   *
   * Creates an instance of an element by specifying the UUID of the element. \
   * This method is used to undo the objects of serialized elements.
   */
  deserializeInstance(elementTypeName: string, parent: ParentUIElement, uuid: string): IUIElement;

  /**
   * Overwrite Element Instance Attributes and Child Elements
   *
   * Copy the element information from `source `to `target`. \
   * Inherited information includes all properties and child elements.
   */
  overwriteInstance(source: IUIElement, target: IUIElement): IUIElement[];

  /**
   * Get an instance of an element from the UUID.
   *
   * This element retrieves the element with the matching UUID from all the elements managed by the controller. \
   * Returns `null` if no matching element exists.
   */
  getInstanceByUUID(uuid: string): IUIElement | null;

  /**
   * Get all instances of child elements belonging to parent element
   *
   * Gets one level of child elements below the specified element. \
   * To recursively retrieve child elements in all hierarchies, use the `getChildrenDeep` method.
   */
  getChildren(parent: ParentUIElement): IUIElement[];

  /**
   * Get all instances of descendant elements belonging to parent element
   *
   * Get recursively the child elements under the specified element. \
   * It takes time to execute because all child elements in all hierarchies are retrieved.
   */
  getChildrenDeep(parent: ParentUIElement): IUIElement[];

  /**
   * Get all instances of child elements belonging to the parent element in the order in which they are placed
   *
   * Gets the child elements under the specified element in the order in which they are drawn on the container. \
   * The `IUIElement#children` property does not guarantee the order, but it does guarantee the order of the child elements returned by this method.
   */
  getSortedChildren(parent: ParentUIElement): IUIElement[];

  /**
   * Get all instances of parent element
   *
   * Get all components of the parent of the specified element. \
   * The retrieved components include elements and containers.
   */
  getParents(child: IUIElement): ParentUIElement[];
}

/* ********************************************************************** *
 * Actions (Common)
 * ********************************************************************** */

interface IUIContainerActionDisplayCore {
  /** Message key identifying body of action */
  readonly messageKey: string;

  /** Parameter definitions */
  readonly parameterDefinition: IUIContainerActionParameterDefinition;
}

interface IUIContainerActionParameterDefinition {
  [key: string]: UIContainerActionParameterDefinitionType;
}

type UIContainerActionParameterDefinitionType = {
  /** Parameter Definition Type */
  readonly type: UIContainerActionParameterCandidateType;

  /** Message key indicating parameter name */
  readonly messageKey: string;

  /** Parameter Candidates */
  readonly candidate?: {
    readonly displayName: string;
    readonly value: string;
  }[];
};

type UIContainerActionParameterCandidateType =
  | 'variable-input' /** constant, input */
  | 'variable-output' /** Variable */
  | 'variable-all' /** Variable, constant, input */
  | 'literal' /** Direct input */
  | 'label-target' /** label definition */
  | 'label' /** Label in Action */
  | 'action' /** Action (except for oneself) */
  | 'page' /** Container page */
  | 'table-select' /** Table */
  | 'checkbox' /** checkbox */
  | 'javascript' /** JavaScript Editor */
  | 'hidden'; /** Hidden Parameters */

interface IUIContainerActionObjectCollector {}

interface ITableSelectActionParameter extends IUIContainerActionParameter {
  /** Gets in the table-column format */
  toColumns(container: IUIContainer): UIContainerActionDialogColumn[];
}

type UIContainerActionDialogColumn = {
  readonly key: string;
  value: string; // Rewritable as it being categorized as a temporary variable
};

/**
 * The action item category name is specified in the message property key of `CAP.Z.IWP.HICHEE.ACTION.ITEM.CATEGORY.NAME. {Any key} `.
 * The category name of action item execution conditions is specified in the message property key of `CAP.Z.IWP.HICHEE.ACTION.CONDITION.CATEGORY.NAME. {Any key} `.
 */
type UIContainerActionCategoryInfo = {
  /** Category ID */
  readonly categoryId: string;

  /** Options */
  readonly options?: UIContainerActionCategoryRegisterOptions;
};

type UIContainerActionCategoryRegisterOptions = {
  /** Sort Number (ascending order) */
  readonly sortNumber?: number;
};

/* ********************************************************************** *
 * Action Chart
 * ********************************************************************** */

interface IActionChartStorage {
  /** Action Chart List */
  readonly charts: IUIContainerActionChart[];

  /** Get Action Chart */
  getChart(uuid: string): IUIContainerActionChart | null;

  /** Create Action Context */
  createContext(chart: IUIContainerActionChart, event: UIContainerActionContextEventType): IUIContainerActionContext;
}

interface IUIContainerActionChart {
  /** UUID */
  readonly uuid: string;

  /** Action Chart Name */
  readonly name: string;

  /** Action item when action is called (normal system) */
  readonly items: IUIContainerActionItem[];

  /** Action item on error occurrence (abnormal system) */
  readonly errorItems: IUIContainerActionItem[];

  /**
   * Sequentially process action items
   *
   * Processes all normal system action items and action execution conditions sequentially. \
   * The action execution condition determination process associated with the action item is executed, and the action item determined to execute the process is executed.
   *
   * Both action items and execution conditions ensure that each is executed one at a time. \
   * Therefore, action items and execution conditions, whether synchronous or asynchronous, are executed in the order in which the array is stored. \
   * If it is implemented asynchronously, it waits for completion.
   *
   * This method is supposed to be executed asynchronously. \
   * Implement `await` or wait process properly when using the method.
   */
  run(context: IUIContainerActionContext, component: IUIComponent): Promise<void>;
}

interface IUIContainerActionContext {
  /** Action Chart */
  readonly chart: IUIContainerActionChart;

  /** Browser Event */
  readonly event: UIContainerActionContextEventType;

  /**
   * Index of the action item to be executed when the action is called
   *
   * Indicates what number of the action items is running in the IUIContainerActionChart#items. \
   * This value is overwritten by IM-BloomMaker itself during the action.
   */
  calledActionIndex: number;

  /**
   * Index of the action item to be performed when an error occurs
   *
   * Indicates what number of the action items is running in the IUIContainerActionChart#errorItems. \
   * This value is overwritten by IM-BloomMaker itself during the action.
   */
  errorOccurredActionIndex: number;
}

type UIContainerActionContextEventType = MouseEvent | KeyboardEvent | null;

/* ********************************************************************** *
 * Action Items
 * ********************************************************************** */

interface IUIContainerActionItem {
  /** UUID */
  readonly uuid: string;

  /** Action Enabled */
  enabled: boolean;

  /** Action Occurrence Condition */
  conditions: IUIContainerActionCondition[];
}

interface IUIContainerActionCondition {}

interface IUIContainerActionItemCore extends IUIContainerActionDisplayCore {
  /**
   * Action Type Name
   *
   * Gets the same string as the class name of the action item.
   */
  readonly actionTypeName: string;

  /**
   * Perform Action Item Process
   *
   * Perform action item processing. \
   * Ensures that `Promise` is returned whether the process is synchronous or asynchronous.
   *
   * This method is supposed to be executed asynchronously. \
   * When using or implementing the method, implement await or wait process properly.
   */
  run(
    context: IUIContainerActionContext,
    container: IUIContainer,
    component: IUIComponent,
    parameters: IUIContainerActionParameterAccessor<IUIContainerActionParameterDefinition>
  ): Promise<void>;

  /**
   * Generate values for dynamically changing parameters
   *
   * Creates values for parameters defined in the action and not displayed on the screen as well as to be complemented on the action side. \
   * You do not need to implement this method if you do not need to complete the parameter values.
   */
  createDynamicParameterValue?(
    container: IUIContainer,
    parameters: IUIContainerActionParameterAccessor<IUIContainerActionParameterDefinition>,
    key: string
  ): string;
}

interface IUIContainerActionItemRepository {
  /** List of Action Item Classes */
  readonly classes: UIContainerActionItemClass[];

  /** Action Item Category List */
  readonly categories: UIContainerActionCategoryInfo[];

  /** Register Action Item Class */
  registerClass(clazz: UIContainerActionItemClass, options?: UIContainerActionRegisterOptions): void;

  /** Register Action Item Category */
  registerCategory(categoryId: string, options?: UIContainerActionCategoryRegisterOptions): void;

  /** Get Action Item Class */
  getClass(actionTypeName: string): UIContainerActionItemClass;

  /** Get Options for Action Item Class */
  getClassOptions(clazz: UIContainerActionItemClass): UIContainerActionRegisterOptions;

  /** Set Action Item Class Options */
  setClassOptions(clazz: UIContainerActionItemClass, options: UIContainerActionRegisterOptions): void;

  /** Get Options for Action Item Category */
  getCategoryOptions(categoryId: string): UIContainerActionCategoryInfo;

  /** Set Action Item Category Options */
  setCategoryOptions(categoryId: string, options: UIContainerActionCategoryRegisterOptions): void;
}

type UIContainerActionItemClass = new () => IUIContainerActionItemCore;

/**
 * The action item execution condition name is specified in the message property key of `CAP.Z.IWP.HICHEE.ACTION.CONDITION.PARAMETER.CONTAINER. {Any key}` .
 * The action item action condition name is specified in the message property key of `CAP.Z.IWP.HICHEE.ACTION.CONDITION.PARAMETER.CONTAINER. {Any key} `.
 */
type UIContainerActionRegisterOptions = {
  /** Category ID */
  readonly categoryId?: string;

  /** Sort Number */
  readonly sortNumber?: number;

  /**
   * Non-display flag in palette
   *
   * This flag is overwritten by the palette in the designer.
   */
  readonly hiddenInPalette?: boolean;
};

/* ********************************************************************** *
 * Action Execution Condition
 * ********************************************************************** */

interface IUIContainerActionConditionRepository {
  /** List of action condition classes */
  readonly classes: UIContainerActionConditionClass[];

  /** List of action conditions categories */
  readonly categories: UIContainerActionCategoryInfo[];

  /** Register Action Condition Class */
  registerClass(clazz: UIContainerActionConditionClass, options?: UIContainerActionRegisterOptions): void;

  /** Register action conditions categories */
  registerCategory(categoryId: string, options?: UIContainerActionCategoryRegisterOptions): void;

  /** Get Action Condition Class */
  getClass(actionTypeName: string): UIContainerActionConditionClass;

  /** Get Options for Action Condition Class */
  getClassOptions(clazz: UIContainerActionConditionClass): UIContainerActionRegisterOptions;

  /** Set Action Condition Class Options */
  setClassOptions(clazz: UIContainerActionConditionClass, options: UIContainerActionRegisterOptions): void;

  /** Get Options for Action Condition Category */
  getCategoryOptions(categoryId: string): UIContainerActionCategoryInfo;

  /** Set Action Condition Category Options */
  setCategoryOptions(categoryId: string, options: UIContainerActionCategoryRegisterOptions): void;
}

interface IUIContainerActionConditionCore extends IUIContainerActionDisplayCore {
  /**
   * Condition Type Name
   *
   * Get the same string as the class name of the action execution condition.
   */
  readonly conditionTypeName: string;

  /**
   * Execute Condition Determination
   *
   * Execute the action execution condition processing and return the judgment result. \
   * Returns `true` if the given condition is met, or `false` if it is not.
   *
   * This method expects asynchronous execution. \
   * When using or implementing a method, implement `await` or wait processing properly.
   */
  matched(
    context: IUIContainerActionContext,
    container: IUIContainer,
    component: IUIComponent,
    parameters: IUIContainerActionParameterAccessor<IUIContainerActionParameterDefinition>
  ): Promise<boolean>;

  /**
   * Generate values for dynamically changing parameters
   *
   * Create values for parameters defined in a condition that are not displayed on the screen but are complemented by the condition. \
   * Implementing this method does not require if you do not need to complete the parameter values.
   */
  createDynamicParameterValue?(
    container: IUIContainer,
    parameters: IUIContainerActionParameterAccessor<IUIContainerActionParameterDefinition>,
    key: string
  ): string;
}

type UIContainerActionConditionClass = new () => IUIContainerActionConditionCore;

/* ********************************************************************** *
 * Action Parameters
 * ********************************************************************** */

interface IUIContainerActionParameterAccessor<DEFINITION> {
  /** Get Properties */
  getParameter<K extends keyof DEFINITION>(key: K): IUIContainerActionParameter;
}

interface IUIContainerActionParameter {
  /** Get at Property Set Value */
  toRaw(): string;

  /** Get at Primitive Value */
  toPrimitive(container: IUIContainer): CSJSPrimitiveAllType;

  /** Get at Argument */
  toArgument(container: IUIContainer): IArgument;
}

/* ********************************************************************** *
 * Parameter
 * ********************************************************************** */

interface IParameterStorage {
  /**
   * Get whether path exists
   *
   * Get whether the path specified relative to the component argument exists.
   */
  existsPath(component: IUIComponent, path: string): boolean;

  /**
   * Read Argument
   *
   * Get the Argument of the variable associated with the specified path relative to the component argument. \
   * Return an Argument indicating null if the path does not exist.
   */
  read(component: IUIComponent, path: string): IArgument;

  /**
   * Read all Arguments, including parent-child relationships
   *
   * Get all the Arguments for the variable associated with the specified path relative to the component argument. \
   * The returned array contains the Arguments in the order they were resolved, following the path's parent.
   */
  readAllChain(component: IUIComponent, path: string): IArgument[];

  /**
   * Write Argument
   *
   * Overwrite the Argument with the variable that is associated with the path specified relative to the component argument. \
   * Return the Argument that was originally stored in the variable associated with the path. \
   * Return an Argument indicating null if the path does not exist.
   */
  writeAsArgument(component: IUIComponent, path: string, argument: IArgument): IArgument;

  /**
   * Write Argument over Parameter
   *
   * Overwrite the Argument with the variable that is associated with the path specified relative to the component argument. \
   * Return the Argument that was originally stored in the variable associated with the path. \
   * Overwrite the parameter of the variable associated with the path with the given Argument type.
   */
  overwriteAsArgument(component: IUIComponent, path: string, argument: IArgument): IArgument;

  /**
   * Write primitive value (No Array Consideration)
   *
   * Overwrite the primitive value to the variable associated with the path specified relative to the component argument. \
   * Primitive values are automatically converted to the corresponding Argument, depending on the type information specified in the variable's parameter.
   */
  writeAsPrimitive(component: IUIComponent, path: string, value: CSJSPrimitiveAllType): IArgument;

  /**
   * Write primitive value (With respect to arrangement)
   *
   * Provide a type conversion equivalent to the writeAsPrimitive method, but it takes an array into account. \
   * If the variable's parameter is an array, it is automatically converted to a 1-element ArrayArgument. \
   * If the variable's parameter is not an array and value is an array, it is converted to Argument for the first element.
   */
  writeAsPrimitiveStrictArrayTyped(component: IUIComponent, path: string, value: CSJSPrimitiveAllType): IArgument;

  /**
   * Convert Argument to Primitive
   *
   * Convert the argument Argument to a primitive value that can be handled by CSJS. \
   * DateArgument converts to a Date class.
   */
  toPrimitive(argument: IArgument): CSJSPrimitiveAllType;

  /**
   * Convert Argument to JSON Object
   *
   * Convert the argument Argument to a JSON object that can be handled by CSJS. \
   * DateArgument converts to a string.
   */
  toJSONObject(argument: IArgument): CSJSPrimitiveAllType;

  /**
   * Convert Argument to Query
   *
   * Convert the argument Argument to a query-parameter format that can be used on the IM-BloomMaker server side.
   */
  toQueryParameters(baseName: string, argument: IArgument): QueryParametersType;

  /**
   * Convert ObjectArgument to a key/string value for use in the query
   *
   * Convert the ObjectArgument argument to a key/string value to create a jQuery compatible query parameter format.
   */
  toQueryKeyValues(argument: IObjectArgument): QueryKeyValuesType;
}

type QueryParametersType = {[key: string]: string};
type QueryKeyValuesType = {[key: string]: string; value: string}[];

/* ********************************************************************** *
 * Argument
 * ********************************************************************** */

interface IArgument {
  /** Argument class */
  readonly clazz: ArgumentClass;

  /** UUID assigned per Argument */
  readonly uuid: string;

  /** Argument type name */
  readonly argumentTypeName: ArgumentTypeName;

  /** Argument Type Display Name */
  readonly displayType: string;

  /** Type display value */
  readonly displayValue: string;

  /** Event Priority */
  readonly eventPriority: number;

  /** Get with string */
  toString(): string;

  /** Get with Numeric Value */
  toNumber(): number;

  /** Get with Boolean Value */
  toBoolean(): boolean;

  /** Array or not */
  isArray(): this is IArrayArgument;

  /** null or not */
  isNull(): boolean;

  /** A valid value (null, not empty) or not */
  isValid(): boolean;

  /** To match or not to match */
  equals(target: IArgument): boolean;

  /** Duplicate Argument */
  clone(): IArgument;
}

interface IPrimitiveArgument extends IArgument {}

interface IChainedArgument extends IPrimitiveArgument {
  /** Argument resolved */
  readonly resolved: IArgument;

  /** Refresh */
  refresh(): void;
}

interface IDateArgument extends IPrimitiveArgument {
  /** Date and time information */
  raw: IDateTime;
}

interface II18nArgument extends IPrimitiveArgument {
  /** Get array of locale ID */
  getLocaleIds(): string[];

  /** Locale ID Existence Check */
  containsLocaleId(localeId: string): boolean;

  /** Get Argument */
  get(localeId: string): string | null;

  /** Configure Argument */
  put(localeId: string, message: string): string | null;

  /** Remove Argument */
  remove(localeId: string): string;
}

interface INumberArgument extends IPrimitiveArgument {}

interface IRelationArgument extends IArgument {
  /** Update Related Argument */
  updateRelatedArgument(oldArgument: IArgument, newArgument: IArgument): void;
}

interface IArrayArgument extends IRelationArgument {
  /** Array Count */
  readonly length: number;

  /** Get if it is empty or not */
  isEmpty(): boolean;

  /** Clear All */
  clear(): void;

  /** Get Argument */
  get(index: number): IArgument;

  /** Configure Argument */
  set(index: number, value: IArgument): IArgument;

  /** Enumerate Argument */
  forEach(callback: ArrayArgumentForEachCallback): void;

  /** Get array with Argument swapped */
  map<T>(callback: ArrayArgumentMapCallback <T>): T[];

  /** Enumerate Argument to see if there is a true return */
  some(callback: ArrayArgumentSomeCallback): boolean;

  /** Add Argument to end */
  push(...values: IArgument[]): void;

  /** Add Argument Content */
  pushAll(value: IArrayArgument): void;

  /** Fetch Argument from the end */
  pop(): IArgument;

  /** Add Argument to top */
  unshift(...value: IArgument[]): void;

  /** Remove Argument from Beginning */
  shift(): IArgument;

  /** Insert Argument */
  insert(index: number, value: IArgument): void;

  /** Insert Argument Content */
  insertAll(index: number, value: IArrayArgument): void;

  /** Remove Argument */
  remove(index: number, size: number): IArgument;

  /** Swap Argument Content */
  replaceAll(value: IArrayArgument): void;
}

type ArrayArgumentForEachCallback = (value: IArgument, index: number, array: IArrayArgument) => void;
type ArrayArgumentMapCallback<T> = (value: IArgument, index: number, array: IArrayArgument) => T;
type ArrayArgumentSomeCallback = (value: IArgument, index: number, array: IArrayArgument) => boolean;

interface IObjectArgument extends IRelationArgument {
  /** Array Count */
  readonly length: number;

  /** Get if it is empty or not */
  isEmpty(): boolean;

  /** Clear All */
  clear(): void;

  /** Get array of keys */
  getKeys(): string[];

  /** key existence check */
  containsKey(key: string): boolean;

  /** Get Argument */
  get(key: string): IArgument;

  /** Configure Argument */
  put(key: string, value: IArgument): IArgument;

  /** Set Argument Content */
  putAll(value: IObjectArgument): void;

  /** Enumerate Argument */
  forEach(callback: ObjectArgumentForEachCallback): void;

  /** Get array with Argument swapped */
  map<T>(callback: ObjectArgumentMapCallback<T>): T[];

  /** Enumerate Argument to see if there is a true return */
  some(callback: ObjectArgumentSomeCallback): boolean;

  /** Remove Argument */
  remove(key: string): IArgument;

  /** Swap Keys */
  replaceKey(sourceKey: string, targetKey: string): void;

  /** Swap Argument Content */
  replaceAll(value: IObjectArgument): void;
}

type ObjectArgumentForEachCallback = (value: IArgument, key: string, object: IObjectArgument) => void;
type ObjectArgumentMapCallback <T> = (value: IArgument, key: string, object: IObjectArgument) => T;
type ObjectArgumentSomeCallback = (value: IArgument, key: string, object: IObjectArgument) => boolean;

interface IPropertyValueArgument extends IChainedArgument {
  /**
   * Raw value that this Argument wraps
   *
   * Use only to get/set raw data for user-specified property values. \
   * Do not use raw to get the property values actually used in the execution screen.
   * Please use the `toString`, `toNumber`, `toBoolean` methods directly.
   */
  raw: IArgument;

  /** Monitor Argument */
  observe(target: IUIComponent, property: IUIComponentProperty): void;

  /** Release Argument */
  release(): void;
}

interface IParameterPathArgument extends IChainedArgument {
  /**
   * Raw value that this Argument wraps
   *
   * Use only to get/set raw data for user-specified property values. \
   * Do not use raw to get the property values actually used in the execution screen.
   * Please use the `toString`, `toNumber`, `toBoolean` methods directly.
   */
  raw: string;

  /** Target Components */
  readonly target: IUIComponent;

  /** Monitor Argument */
  observe(): void;

  /** Release Argument */
  release(): void;
}

interface IDateTime {
  /** Datetime object */
  readonly date: Date;

  /** Time difference (ms) */
  readonly offset: number;

  /** Strings */
  toString(): string;

  /** Quantify (epoch millisecond) */
  toNumber(): number;

  /** String in ISO8601 format */
  toISOString(offset: number): string;

  /** Component Division */
  getComponents(): DateTimeComponents;

  /** Duplicate Instance */
  clone(): IDateTime;

  /** Match Comparison */
  equals(target: IDateTime): boolean;

  /** Differential Comparison */
  compare(target: IDateTime): number;
}

type DateTimeComponents = {
  /** Date */
  readonly date: {
    readonly iso: string;
    readonly year: number;
    readonly month: number; // 1-12
    readonly day: number;
  };

  /** Time */
  readonly time: {
    readonly iso: string;
    readonly hours: number;
    readonly minutes: number;
    readonly seconds: number;
    readonly milliseconds: number;
  };

  /** Time difference */
  readonly offset: {
    readonly iso: string;
    readonly milliseconds: number;
  };
};

interface IFraction {
  /** Molecules */
  readonly numerator: number;

  /** Denominator */
  readonly denominator: number;

  /** Positive number or not */
  readonly plus: boolean;

  /** 0 or not */
  readonly zero: boolean;

  /** Negative number or not */
  readonly minus: boolean;

  /** Whether it is divisible or not */
  readonly divisible: boolean;

  /** String */
  toString(): string;

  /** Quantify (accuracy reduction attention) */
  toNumber(): number;

  /** Duplicate Instance */
  clone(): IFraction;

  /** Match Comparison */
  equals(target: IFraction): boolean;

  /** Differential Comparison */
  compare(target: IFraction): number;

  /** Add */
  add(num: number | string | IFraction, denom?: number): IFraction;

  /** Subtract */
  sub(num: number | string | IFraction, denom?: number): IFraction;

  /** Multiplication */
  mul(num: number | string | IFraction, denom?: number): IFraction;

  /** Division */
  div(num: number | string | IFraction, denom?: number): IFraction;

  /** Remainder */
  mod(num: number | string | IFraction, denom?: number): IFraction;

  /** Absolute value */
  abs(): IFraction;

  /** Truncate */
  floor(): number;

  /** Round */
  round(): number;

  /** Advance */
  ceil(): number;
}

interface IKeyValueMap<K extends string, V> {
  /** Number of Keys */
  readonly length: number;

  /** Array of keys */
  readonly keys: K[];

  /** Array of values */
  readonly values: V[];

  /** Clear */
  clear(): void;

  /** Duplicate Instance */
  clone(): IKeyValueMap<K, V>;

  /** key or empty */
  isEmpty(): boolean;

  /** Repeat */
  each(callback: (key: K, value: V) => boolean | void): number;

  /** Fetch only elements that return true */
  filter(callback: (key: K, value: V) => boolean): IKeyValueMap<K, V>;

  /** contains a key */
  containsKey(key: K): boolean;

  /** Whether it contains a value or not */
  containsValue(value: V): boolean;

  /** Get Value */
  get(key: K): V;

  /** Set Value */
  put(key: K, value: V): V;

  /** Delete Key */
  remove(key: K): V;
}

type ArgumentClass =
  | 'ArrayArgument' /** Array */
  | 'BooleanArgument' /** Boolean value */
  | 'DateArgument' /** Date */
  | 'DoubleArgument' /** Float */
  | 'FormulaArgument' /** Expression */
  | 'FractionArgument' /** High precision decimal */
  | 'I18nArgument' /** Multilingual */
  | 'IntegerArgument' /** Integer */
  | 'ObjectArgument' /** Object/Hash Map */
  | 'NullArgument' /** null */
  | 'ParameterPathArgument' /** Variable Path */
  | 'PropertyValueArgument' /** Property Value */
  | 'StringArgument'; /** string */

type ArgumentTypeName =
  | 'array' /** Array */
  | 'boolean' /** Boolean value */
  | 'date' /** Date */
  | 'double' /** Float */
  | 'formula' /** Expression */
  | 'fraction' /** High precision decimal */
  | 'im-repository' /** IM-Repository definition */
  | 'i18n' /** Multilingual */
  | 'integer' /** Integer */
  | 'object' /** Object/Hash Map */
  | 'null' /** null */
  | 'parameter-path' /** Variable path */
  | 'property-value' /** Property value */
  | 'string'; /** String */

/* ********************************************************************** *
 * HTMLElementBuilder
 * ********************************************************************** */

interface IHTMLElementBuilder {
  /** Class of HTML Element Builder */
  readonly clazz: HTMLElementBuilderClass;

  /** HTML tag */
  readonly tag: HTMLElement;

  /** Configure of the HTML Element Builder */
  readonly config: IHTMLElementConfig;

  /** Get if class name is kept */
  hasClass(className: string): boolean;

  /** Add Class Name */
  addClass(className: string): IHTMLElementBuilder;

  /** Delete Class Name */
  removeClass(className: string): IHTMLElementBuilder;

  /** Clear Class Name */
  clearClass(): IHTMLElementBuilder;

  /** Add/remove class name with argument on */
  switchClass(on: boolean, className: string): IHTMLElementBuilder;

  /** Get CSS */
  getCSS(name: string): string | null;

  /** Configure CSS */
  setCSS(name: string, value: string): IHTMLElementBuilder;

  /** Remove CSS */
  removeCSS(name: string): IHTMLElementBuilder;

  /** Set/Unset CSS by argument on */
  switchCSS(on: boolean, name: string, value: string): IHTMLElementBuilder;

  /** CSS configuration shorthand to specify location */
  setPosition(position: {top?: number; right?: number; bottom?: number; left?: number}): IHTMLElementBuilder;

  /** CSS configuration shorthand to specify size */
  setSize(size: {width?: string | number; height?: string | number}): IHTMLElementBuilder;

  /** Get Attributes */
  getAttribute(name: string): string | null;

  /**
   * Set Attributes
   *
   * To change the `'style' ', use the `setCSS` method. \
   * To change the `'class' ', use the `addClass` method.
   */
  setAttribute(name: string, value?: string): IHTMLElementBuilder;

  /**
   * Delete Attribute
   *
   * To delete the `'style' ', use the `removeCSS` method. \
   * To delete the `'class' ', use the `clearClass` method.
   */
  removeAttribute(name: string): IHTMLElementBuilder;

  /** Set/Unset attribute by argument on */
  switchAttribute(on: boolean, name: string, value?: string): IHTMLElementBuilder;

  /**
   * Get Tag Value
   *
   * Get the value attribute value of the tag managed by this Element Builder.
   */
  getValue(): string | null;

  /**
   * Set Tag Value
   *
   * Set the value attribute value of the tag managed by this element builder.
   */
  setValue(value: string | null): IHTMLElementBuilder;

  /**
   * Get Text in Tag
   *
   * Get the string contained in the text node in the tags managed by this Element Builder.
   */
  getText(): string | null;

  /**
   * Set Text in Tag
   *
   * Set the text node for the tags managed by this element builder. \
   * Add a text node to the end of the child tag (After Trailing Group).
   */
  setText(text: string | null): IHTMLElementBuilder;

  /**
   * Display Tags
   *
   * Add onscreen styles to the tags managed by this Element Builder.
   */
  show(): IHTMLElementBuilder;

  /**
   * Hide Tags
   *
   * Add a style that does not appear on the screen to the tags managed by this Element Builder.
   */
  hide(): IHTMLElementBuilder;

  /**
   * Add Child Tag to the First Group
   *
   * Add child tags to the first group of child tags in the order in which you called the methods. \
   * When building HTML tags with the build method, add child tags in the order of top group + bottom group.
   */
  prependChild(value: HTMLElement | IHTMLElementBuilder | IUIElement): IHTMLElementBuilder;

  /**
   * Add child tag to bottom group
   *
   * Add child tags to the bottom group of child tags in the order in which you called the methods. \
   * When building HTML tags with the build method, add child tags in the order of top group + bottom group.
   */
  appendChild(value: HTMLElement | IHTMLElementBuilder | IUIElement): IHTMLElementBuilder;

  /**
   * Delete Child Tag
   *
   * prependChild, child tags added in appendChild method,
   * Or, delete the tag specifiedby the argument from the child tags that were already added when this Element Builder was initialized.
   */
  removeChild(value: HTMLElement | IHTMLElementBuilder | IUIElement): IHTMLElementBuilder;

  /**
   * Delete all child tags
   *
   * prependChild, child tags added in appendChild method,
   * Or, delete all child tags that were already added when this element builder was initialized.
   */
  removeChildren(): IHTMLElementBuilder;

  /**
   * Build HTML Tags
   *
   * Apply any changes made to the tag in each method.
   */
  build(): IHTMLElementBuilder;

  /**
   * Reset Internal Information
   *
   * Discard changes made to the tag in each method.
   */
  reset(): IHTMLElementBuilder;
}

type HTMLElementBuilderClass = 'HTMLElementBuilder';

type IHTMLElementConfig = {
  /** Class List */
  readonly classNames: string[];

  /** Tag Attribute */
  readonly attributes: {
    [key: string]: string | null;
  };

  /** New value */
  readonly newValue: string | null;

  /** Old Text */
  readonly oldText: string | null;

  /** Changed Text */
  readonly newText: string | null;

  /** HTML element added to top group */
  readonly reservePrepend: HTMLElement[];

  /** HTML element added to the bottom group */
  readonly reserveAppend: HTMLElement[];

  /** Deleted HTML element */
  readonly reserveRemove: HTMLElement[];

  /** Added HTML Element Builder */
  readonly builders: IHTMLElementBuilder[];
};

/* ********************************************************************** *
 * BrowserHelper
 * ********************************************************************** */

type BrowserInformation = {os: OperatingSystem; application: BrowserApplication};

type OperatingSystem = 'Unknown' /** Unknown */ | 'Windows' /** Windows */ | 'Mac'; /** Mac */
type BrowserApplication =
  | 'Unknown' /** Unknown */
  | 'MSIE' /** IE */
  | 'MS-Edge' /** MS-Edge */
  | 'Chromium-Edge' /** Chromium-Edge */
  | 'Firefox' /** Firefox */
  | 'Chrome' /** Chrome */
  | 'Safari'; /** Safari */

/* ********************************************************************** *
 * MessageBuilder
 * ********************************************************************** */

interface IMessageBuilder {
  /** Get Message */
  get(key: string, ...args: MessageBuilderArgumentType): string;

  /** Get Message */
  getLocalize(key: string, localeId: string, ...args: MessageBuilderArgumentType): string;

  /** Register Message */
  register(uuid: string): void;
}

type MessageBuilderArgumentType = (string | number | boolean)[];

/* ********************************************************************** *
 * JavaScript compatibility
 * ********************************************************************** */

type CSJSPrimitiveSingleType = string | number | boolean | object | null | undefined;
type CSJSPrimitiveAllType = CSJSPrimitiveSingleType | CSJSPrimitiveSingleType[];
